from typing import Type


def register_raw_df_type(df_type: Type) -> None:  # pragma: no cover
    """TODO: This function is to be removed before 0.9.0

    .. deprecated:: 0.8.0
        Register using :func:`fugue.api.is_df` instead.
    """
    raise DeprecationWarning("use fugue.api.is_df to register the dataframe")



from typing import Any, Dict, Iterable, List, Optional, Tuple

import dask.dataframe as dd
import pandas
import pyarrow as pa
from triad.collections.schema import Schema
from triad.utils.assertion import assert_arg_not_none, assert_or_throw
from triad.utils.pyarrow import to_pandas_dtype

from fugue.dataframe import (
    ArrowDataFrame,
    DataFrame,
    LocalBoundedDataFrame,
    PandasDataFrame,
)
from fugue.dataframe.dataframe import _input_schema
from fugue.exceptions import FugueDataFrameOperationError
from fugue.plugins import (
    as_local_bounded,
    count,
    drop_columns,
    get_column_names,
    get_num_partitions,
    head,
    is_bounded,
    is_df,
    is_empty,
    is_local,
    rename,
    select_columns,
)

from ._utils import DASK_UTILS, get_default_partitions


class DaskDataFrame(DataFrame):
    """DataFrame that wraps Dask DataFrame. Please also read
    |DataFrameTutorial| to understand this Fugue concept

    :param df: :class:`dask:dask.dataframe.DataFrame`,
      pandas DataFrame or list or iterable of arrays
    :param schema: |SchemaLikeObject| or :class:`spark:pyspark.sql.types.StructType`,
      defaults to None.
    :param num_partitions: initial number of partitions for the dask dataframe
      defaults to 0 to get the value from `fugue.dask.default.partitions`
    :param type_safe: whether to cast input data to ensure type safe, defaults to True

    .. note::

        For :class:`dask:dask.dataframe.DataFrame`, schema must be None
    """

    def __init__(  # noqa: C901
        self,
        df: Any = None,
        schema: Any = None,
        num_partitions: int = 0,
        type_safe=True,
    ):
        if num_partitions <= 0:
            num_partitions = get_default_partitions()
        if df is None:
            schema = _input_schema(schema).assert_not_empty()
            df = []
        if isinstance(df, DaskDataFrame):
            super().__init__(df.schema)
            self._native: dd.DataFrame = df._native
            return
        elif isinstance(df, (dd.DataFrame, dd.Series)):
            if isinstance(df, dd.Series):
                df = df.to_frame()
            pdf = df
            schema = None if schema is None else _input_schema(schema)
        elif isinstance(df, (pandas.DataFrame, pandas.Series)):
            if isinstance(df, pandas.Series):
                df = df.to_frame()
            pdf = dd.from_pandas(df, npartitions=num_partitions, sort=False)
            schema = None if schema is None else _input_schema(schema)
        elif isinstance(df, Iterable):
            schema = _input_schema(schema).assert_not_empty()
            t = PandasDataFrame(df, schema)
            pdf = dd.from_pandas(t.native, npartitions=num_partitions, sort=False)
            type_safe = False
        else:
            raise ValueError(f"{df} is incompatible with DaskDataFrame")
        pdf, schema = self._apply_schema(pdf, schema, type_safe)
        super().__init__(schema)
        self._native = pdf

    @property
    def native(self) -> dd.DataFrame:
        """The wrapped Dask DataFrame"""
        return self._native

    def native_as_df(self) -> dd.DataFrame:
        return self._native

    @property
    def is_local(self) -> bool:
        return False

    def as_local_bounded(self) -> LocalBoundedDataFrame:
        res = PandasDataFrame(self.as_pandas(), self.schema)
        if self.has_metadata:
            res.reset_metadata(self.metadata)
        return res

    @property
    def is_bounded(self) -> bool:
        return True

    @property
    def empty(self) -> bool:
        return DASK_UTILS.empty(self.native)

    @property
    def num_partitions(self) -> int:
        return _dd_get_num_partitions(self.native)

    def _drop_cols(self, cols: List[str]) -> DataFrame:
        cols = (self.schema - cols).names
        return self._select_cols(cols)

    def _select_cols(self, cols: List[Any]) -> DataFrame:
        schema = self.schema.extract(cols)
        return DaskDataFrame(self.native[schema.names], schema, type_safe=False)

    def peek_array(self) -> List[Any]:
        self.assert_not_empty()
        return self.as_pandas().iloc[0].values.tolist()

    def persist(self, **kwargs: Any) -> "DaskDataFrame":
        self._native = self.native.persist(**kwargs)
        self._native.count().compute()
        return self

    def count(self) -> int:
        return self.native.shape[0].compute()

    def as_pandas(self) -> pandas.DataFrame:
        return self.native.compute().reset_index(drop=True)

    def rename(self, columns: Dict[str, str]) -> DataFrame:
        try:
            schema = self.schema.rename(columns)
        except Exception as e:
            raise FugueDataFrameOperationError from e
        df = self.native.rename(columns=columns)
        return DaskDataFrame(df, schema, type_safe=False)

    def alter_columns(self, columns: Any) -> DataFrame:
        new_schema = self._get_altered_schema(columns)
        if new_schema == self.schema:
            return self
        new_pdf = self.native.assign()
        pd_types = to_pandas_dtype(new_schema.pa_schema)
        for k, v in new_schema.items():
            if not v.type.equals(self.schema[k].type):
                old_type = self.schema[k].type
                new_type = v.type
                # int -> str
                if pa.types.is_integer(old_type) and pa.types.is_string(new_type):
                    series = new_pdf[k]
                    ns = series.isnull()
                    series = series.fillna(0).astype(int).astype(str)
                    new_pdf[k] = series.mask(ns, None)
                # bool -> str
                elif pa.types.is_boolean(old_type) and pa.types.is_string(new_type):
                    series = new_pdf[k]
                    ns = series.isnull()
                    positive = series != 0
                    new_pdf[k] = "False"
                    new_pdf[k] = new_pdf[k].mask(positive, "True").mask(ns, None)
                # str -> bool
                elif pa.types.is_string(old_type) and pa.types.is_boolean(new_type):
                    series = new_pdf[k]
                    ns = series.isnull()
                    new_pdf[k] = (
                        series.fillna("true")
                        .apply(lambda x: None if x is None else x.lower())
                        .mask(ns, None)
                    )
                elif pa.types.is_integer(new_type):
                    series = new_pdf[k]
                    ns = series.isnull()
                    series = series.fillna(0).astype(pd_types[k])
                    new_pdf[k] = series.mask(ns, None)
                else:
                    series = new_pdf[k]
                    ns = series.isnull()
                    series = series.astype(pd_types[k])
                    new_pdf[k] = series.mask(ns, None)
        return DaskDataFrame(new_pdf, new_schema, type_safe=True)

    def as_array(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> List[Any]:
        df: DataFrame = self
        if columns is not None:
            df = df[columns]
        return ArrowDataFrame(df.as_pandas(), schema=df.schema).as_array(
            type_safe=type_safe
        )

    def as_array_iterable(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> Iterable[Any]:
        yield from self.as_array(columns=columns, type_safe=type_safe)

    def head(
        self, n: int, columns: Optional[List[str]] = None
    ) -> LocalBoundedDataFrame:
        ddf = self.native if columns is None else self.native[columns]
        schema = self.schema if columns is None else self.schema.extract(columns)
        return PandasDataFrame(ddf.head(n, compute=True, npartitions=-1), schema=schema)

    def _apply_schema(
        self, pdf: dd.DataFrame, schema: Optional[Schema], type_safe: bool = True
    ) -> Tuple[dd.DataFrame, Schema]:
        if not type_safe:
            assert_arg_not_none(pdf, "pdf")
            assert_arg_not_none(schema, "schema")
            return pdf, schema
        DASK_UTILS.ensure_compatible(pdf)
        if pdf.columns.dtype == "object":  # pdf has named schema
            pschema = Schema(DASK_UTILS.to_schema(pdf))
            if schema is None or pschema == schema:
                return pdf, pschema.assert_not_empty()
            pdf = pdf[schema.assert_not_empty().names]
        else:  # pdf has no named schema
            schema = _input_schema(schema).assert_not_empty()
            assert_or_throw(
                pdf.shape[1] == len(schema),
                lambda: ValueError(
                    f"Pandas datafame column count doesn't match {schema}"
                ),
            )
            pdf.columns = schema.names
        return DASK_UTILS.enforce_type(pdf, schema.pa_schema, null_safe=True), schema


@is_df.candidate(lambda df: isinstance(df, dd.DataFrame))
def _dd_is_df(df: dd.DataFrame) -> bool:
    return True


@get_num_partitions.candidate(lambda df: isinstance(df, dd.DataFrame))
def _dd_get_num_partitions(df: dd.DataFrame) -> int:
    return df.npartitions


@count.candidate(lambda df: isinstance(df, dd.DataFrame))
def _dd_count(df: dd.DataFrame) -> int:
    return df.shape[0].compute()


@is_bounded.candidate(lambda df: isinstance(df, dd.DataFrame))
def _dd_is_bounded(df: dd.DataFrame) -> bool:
    return True


@is_empty.candidate(lambda df: isinstance(df, dd.DataFrame))
def _dd_is_empty(df: dd.DataFrame) -> bool:
    return DASK_UTILS.empty(df)


@is_local.candidate(lambda df: isinstance(df, dd.DataFrame))
def _dd_is_local(df: dd.DataFrame) -> bool:
    return False


@as_local_bounded.candidate(lambda df: isinstance(df, dd.DataFrame))
def _dd_as_local(df: dd.DataFrame) -> bool:
    return df.compute()


@get_column_names.candidate(lambda df: isinstance(df, dd.DataFrame))
def _get_dask_dataframe_columns(df: dd.DataFrame) -> List[Any]:
    return list(df.columns)


@rename.candidate(lambda df, *args, **kwargs: isinstance(df, dd.DataFrame))
def _rename_dask_dataframe(df: dd.DataFrame, columns: Dict[str, Any]) -> dd.DataFrame:
    if len(columns) == 0:
        return df
    _assert_no_missing(df, columns.keys())
    return df.rename(columns=columns)


@drop_columns.candidate(lambda df, *args, **kwargs: isinstance(df, dd.DataFrame))
def _drop_dd_columns(
    df: dd.DataFrame, columns: List[str], as_fugue: bool = False
) -> Any:
    cols = [x for x in df.columns if x not in columns]
    if len(cols) == 0:
        raise FugueDataFrameOperationError("cannot drop all columns")
    if len(cols) + len(columns) != len(df.columns):
        _assert_no_missing(df, columns)
    return _adjust_df(df[cols], as_fugue=as_fugue)


@select_columns.candidate(lambda df, *args, **kwargs: isinstance(df, dd.DataFrame))
def _select_dd_columns(
    df: dd.DataFrame, columns: List[Any], as_fugue: bool = False
) -> Any:
    if len(columns) == 0:
        raise FugueDataFrameOperationError("must select at least one column")
    _assert_no_missing(df, columns)
    return _adjust_df(df[columns], as_fugue=as_fugue)


@head.candidate(lambda df, *args, **kwargs: isinstance(df, dd.DataFrame))
def _dd_head(
    df: dd.DataFrame,
    n: int,
    columns: Optional[List[str]] = None,
    as_fugue: bool = False,
) -> pandas.DataFrame:
    if columns is not None:
        df = df[columns]
    res = df.head(n, compute=True, npartitions=-1)
    return PandasDataFrame(res) if as_fugue else res


def _assert_no_missing(df: dd.DataFrame, columns: Iterable[Any]) -> None:
    missing = set(columns) - set(df.columns)
    if len(missing) > 0:
        raise FugueDataFrameOperationError("found nonexistent columns: {missing}")


def _adjust_df(res: dd.DataFrame, as_fugue: bool):
    return res if not as_fugue else DaskDataFrame(res)



from typing import Any

import dask.dataframe as dd
from dask.distributed import Client
from triad import run_at_def

from fugue import DataFrame, register_execution_engine
from fugue.dev import (
    DataFrameParam,
    ExecutionEngineParam,
    fugue_annotated_param,
    is_pandas_or,
)
from fugue.plugins import as_fugue_dataset, infer_execution_engine
from fugue_dask._utils import DASK_UTILS
from fugue_dask.dataframe import DaskDataFrame
from fugue_dask.execution_engine import DaskExecutionEngine


@infer_execution_engine.candidate(
    lambda objs: is_pandas_or(objs, (dd.DataFrame, DaskDataFrame))
)
def _infer_dask_client(objs: Any) -> Any:
    return DASK_UTILS.get_or_create_client()


@as_fugue_dataset.candidate(lambda df, **kwargs: isinstance(df, dd.DataFrame))
def _dask_as_fugue_df(df: dd.DataFrame, **kwargs: Any) -> DaskDataFrame:
    return DaskDataFrame(df, **kwargs)


def _register_engines() -> None:
    register_execution_engine(
        "dask",
        lambda conf, **kwargs: DaskExecutionEngine(conf=conf),
        on_dup="ignore",
    )
    register_execution_engine(
        Client,
        lambda engine, conf, **kwargs: DaskExecutionEngine(
            dask_client=engine, conf=conf
        ),
        on_dup="ignore",
    )


@fugue_annotated_param(DaskExecutionEngine)
class _DaskExecutionEngineParam(ExecutionEngineParam):
    pass


@fugue_annotated_param(dd.DataFrame)
class _DaskDataFrameParam(DataFrameParam):
    def to_input_data(self, df: DataFrame, ctx: Any) -> Any:
        assert isinstance(ctx, DaskExecutionEngine)
        return ctx.to_df(df).native

    def to_output_df(self, output: Any, schema: Any, ctx: Any) -> DataFrame:
        assert isinstance(output, dd.DataFrame)
        assert isinstance(ctx, DaskExecutionEngine)
        return ctx.to_df(output, schema=schema)

    def count(self, df: DataFrame) -> int:  # pragma: no cover
        raise NotImplementedError("not allowed")


@run_at_def
def _register() -> None:
    """Register Dask Execution Engine

    .. note::

        This function is automatically called when you do

        >>> import fugue_dask
    """
    _register_engines()



# flake8: noqa
from .api import *
from .execution_engine import AnyExecutionEngine, ExecutionEngine, MapEngine, SQLEngine
from .factory import (
    infer_execution_engine,
    make_execution_engine,
    make_sql_engine,
    register_default_execution_engine,
    register_default_sql_engine,
    register_execution_engine,
    register_sql_engine,
)
from .native_execution_engine import NativeExecutionEngine, QPDPandasEngine



import os
import pathlib
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union
from urllib.parse import urlparse

import fs as pfs
import pandas as pd
from triad.collections.dict import ParamDict
from triad.collections.fs import FileSystem
from triad.collections.schema import Schema
from triad.utils.assertion import assert_or_throw

from fugue.dataframe import LocalBoundedDataFrame, LocalDataFrame, PandasDataFrame


class FileParser(object):
    def __init__(self, path: str, format_hint: Optional[str] = None):
        last = len(path)
        has_glob = False
        self._orig_format_hint = format_hint
        for i in range(len(path)):
            if path[i] in ["/", "\\"]:
                last = i
            if path[i] in ["*", "?"]:
                has_glob = True
                break
        if not has_glob:
            self._uri = urlparse(path)
            self._glob_pattern = ""
            self._path = self._uri.path
        else:
            self._uri = urlparse(path[:last])
            self._glob_pattern = path[last + 1 :]
            self._path = pfs.path.combine(self._uri.path, self._glob_pattern)

        if format_hint is None or format_hint == "":
            for k, v in _FORMAT_MAP.items():
                if self.suffix.endswith(k):
                    self._format = v
                    return
            raise NotImplementedError(f"{self.suffix} is not supported")
        else:
            assert_or_throw(
                format_hint in _FORMAT_MAP.values(),
                NotImplementedError(f"{format_hint} is not supported"),
            )
            self._format = format_hint

    def assert_no_glob(self) -> "FileParser":
        assert_or_throw(self.glob_pattern == "", f"{self.path} has glob pattern")
        return self

    def with_glob(self, glob: str, format_hint: Optional[str] = None) -> "FileParser":
        uri = self.uri
        if glob != "":
            uri = pfs.path.combine(uri, glob)
        return FileParser(uri, format_hint or self._orig_format_hint)

    @property
    def glob_pattern(self) -> str:
        return self._glob_pattern

    @property
    def uri(self) -> str:
        return self._uri.geturl()

    @property
    def uri_with_glob(self) -> str:
        if self.glob_pattern == "":
            return self.uri
        return pfs.path.combine(self.uri, self.glob_pattern)

    @property
    def parent(self) -> str:
        dn = os.path.dirname(self.uri)
        return dn if dn != "" else "."

    @property
    def scheme(self) -> str:
        return self._uri.scheme

    @property
    def path(self) -> str:
        return self._path

    @property
    def suffix(self) -> str:
        return "".join(pathlib.Path(self.path.lower()).suffixes)

    @property
    def file_format(self) -> str:
        return self._format


def load_df(
    uri: Union[str, List[str]],
    format_hint: Optional[str] = None,
    columns: Any = None,
    fs: Optional[FileSystem] = None,
    **kwargs: Any,
) -> LocalBoundedDataFrame:
    if isinstance(uri, str):
        fp = [FileParser(uri, format_hint)]
    else:
        fp = [FileParser(u, format_hint) for u in uri]
    dfs: List[pd.DataFrame] = []
    schema: Any = None
    for f in _get_single_files(fp, fs):
        df, schema = _FORMAT_LOAD[f.file_format](f.assert_no_glob(), columns, **kwargs)
        dfs.append(df)
    return PandasDataFrame(pd.concat(dfs), schema)


def save_df(
    df: LocalDataFrame,
    uri: str,
    format_hint: Optional[str] = None,
    mode: str = "overwrite",
    fs: Optional[FileSystem] = None,
    **kwargs: Any,
) -> None:
    assert_or_throw(
        mode in ["overwrite", "error"], NotImplementedError(f"{mode} is not supported")
    )
    p = FileParser(uri, format_hint).assert_no_glob()
    if fs is None:
        fs = FileSystem()
    if fs.exists(uri):
        assert_or_throw(mode == "overwrite", FileExistsError(uri))
        try:
            fs.remove(uri)
        except Exception:
            try:
                fs.removetree(uri)
            except Exception:  # pragma: no cover
                pass
    _FORMAT_SAVE[p.file_format](df, p, **kwargs)


def _get_single_files(
    fp: Iterable[FileParser], fs: Optional[FileSystem]
) -> Iterable[FileParser]:
    if fs is None:
        fs = FileSystem()
    for f in fp:
        if f.glob_pattern != "":
            files = [
                FileParser(pfs.path.combine(f.uri, pfs.path.basename(x.path)))
                for x in fs.opendir(f.uri).glob(f.glob_pattern)
            ]
            yield from _get_single_files(files, fs)
        else:
            yield f


def _save_parquet(df: LocalDataFrame, p: FileParser, **kwargs: Any) -> None:
    df.as_pandas().to_parquet(
        p.uri, **{"engine": "pyarrow", "schema": df.schema.pa_schema, **kwargs}
    )


def _load_parquet(
    p: FileParser, columns: Any = None, **kwargs: Any
) -> Tuple[pd.DataFrame, Any]:
    if columns is None:
        pdf = pd.read_parquet(p.uri, **{"engine": "pyarrow", **kwargs})
        return pdf, None
    if isinstance(columns, list):  # column names
        pdf = pd.read_parquet(p.uri, columns=columns, **{"engine": "pyarrow", **kwargs})
        return pdf, None
    schema = Schema(columns)
    pdf = pd.read_parquet(
        p.uri, columns=schema.names, **{"engine": "pyarrow", **kwargs}
    )
    return pdf, schema


def _save_csv(df: LocalDataFrame, p: FileParser, **kwargs: Any) -> None:
    df.as_pandas().to_csv(p.uri, **{"index": False, "header": False, **kwargs})


def _safe_load_csv(path: str, **kwargs: Any) -> pd.DataFrame:
    def load_dir() -> pd.DataFrame:
        fs = FileSystem()
        return pd.concat(
            [
                pd.read_csv(pfs.path.combine(path, pfs.path.basename(x.path)), **kwargs)
                for x in fs.opendir(path).glob("*.csv")
            ]
        )

    try:
        return pd.read_csv(path, **kwargs)
    except IsADirectoryError:
        return load_dir()
    except pd.errors.ParserError:  # pragma: no cover
        # for python < 3.7
        return load_dir()
    except PermissionError:  # pragma: no cover
        # for windows
        return load_dir()


def _load_csv(  # noqa: C901
    p: FileParser, columns: Any = None, **kwargs: Any
) -> Tuple[pd.DataFrame, Any]:
    kw = ParamDict(kwargs)
    infer_schema = kw.get("infer_schema", False)
    if infer_schema and columns is not None and not isinstance(columns, list):
        raise ValueError("can't set columns as a schema when infer schema is true")
    if not infer_schema:
        kw["dtype"] = object
    if "infer_schema" in kw:
        del kw["infer_schema"]
    header: Any = False
    if "header" in kw:
        header = kw["header"]
        del kw["header"]
    if str(header) in ["True", "0"]:
        pdf = _safe_load_csv(p.uri, **{"index_col": False, "header": 0, **kw})
        if columns is None:
            return pdf, None
        if isinstance(columns, list):  # column names
            return pdf[columns], None
        schema = Schema(columns)
        return pdf[schema.names], schema
    if header is None or str(header) == "False":
        if columns is None:
            raise ValueError("columns must be set if without header")
        if isinstance(columns, list):  # column names
            pdf = _safe_load_csv(
                p.uri, **{"index_col": False, "header": None, "names": columns, **kw}
            )
            return pdf, None
        schema = Schema(columns)
        pdf = _safe_load_csv(
            p.uri, **{"index_col": False, "header": None, "names": schema.names, **kw}
        )
        return pdf, schema
    else:
        raise NotImplementedError(f"{header} is not supported")


def _save_json(df: LocalDataFrame, p: FileParser, **kwargs: Any) -> None:
    df.as_pandas().to_json(p.uri, **{"orient": "records", "lines": True, **kwargs})


def _safe_load_json(path: str, **kwargs: Any) -> pd.DataFrame:
    kw = {"orient": "records", "lines": True, **kwargs}
    try:
        return pd.read_json(path, **kw)
    except (IsADirectoryError, PermissionError):
        fs = FileSystem()
        return pd.concat(
            [
                pd.read_json(pfs.path.combine(path, pfs.path.basename(x.path)), **kw)
                for x in fs.opendir(path).glob("*.json")
            ]
        )


def _load_json(
    p: FileParser, columns: Any = None, **kwargs: Any
) -> Tuple[pd.DataFrame, Any]:
    pdf = _safe_load_json(p.uri, **kwargs).reset_index(drop=True)
    if columns is None:
        return pdf, None
    if isinstance(columns, list):  # column names
        return pdf[columns], None
    schema = Schema(columns)
    return pdf[schema.names], schema


_FORMAT_MAP: Dict[str, str] = {
    ".csv": "csv",
    ".csv.gz": "csv",
    ".parquet": "parquet",
    ".json": "json",
    ".json.gz": "json",
}

_FORMAT_LOAD: Dict[str, Callable[..., Tuple[pd.DataFrame, Any]]] = {
    "csv": _load_csv,
    "parquet": _load_parquet,
    "json": _load_json,
}

_FORMAT_SAVE: Dict[str, Callable] = {
    "csv": _save_csv,
    "parquet": _save_parquet,
    "json": _save_json,
}



import polars as pl
from triad import Schema

from fugue.dataframe.arrow_dataframe import _build_empty_arrow


def build_empty_pl(schema: Schema) -> pl.DataFrame:
    return pl.from_arrow(_build_empty_arrow(schema))



import pickle
from typing import Any, Dict, List, Optional, Tuple

import pandas as pd
import pyarrow as pa
import ray
import ray.data as rd
from triad import Schema

from fugue.dataframe.arrow_dataframe import _build_empty_arrow

from .._constants import _ZERO_COPY

_RAY_NULL_REPR = "__RAY_NULL__"


def is_materialized(df: rd.Dataset) -> bool:
    if hasattr(rd.dataset, "MaterializedDataset"):
        return isinstance(df, rd.dataset.MaterializedDataset)
    return df.is_fully_executed()  # pragma: no cover


def materialize(df: rd.Dataset) -> rd.Dataset:
    if not is_materialized(df):
        if hasattr(df, "materialize"):
            df = df.materialize()
        else:  # pragma: no cover
            df = df.fully_executed()
    return df


def get_dataset_format(df: rd.Dataset) -> Tuple[Optional[str], rd.Dataset]:
    df = materialize(df)
    if df.count() == 0:
        return None, df
    if ray.__version__ < "2.5.0":  # pragma: no cover
        if hasattr(df, "_dataset_format"):  # pragma: no cover
            return df._dataset_format(), df  # ray<2.2
        ctx = rd.context.DatasetContext.get_current()
        ctx.use_streaming_executor = False
        return df.dataset_format(), df  # ray>=2.2
    else:
        schema = df.schema(fetch_if_missing=True)
        if schema is None:  # pragma: no cover
            return None, df
        if isinstance(schema.base_schema, pa.Schema):
            return "arrow", df
        return "pandas", df


def to_schema(schema: Any) -> Schema:  # pragma: no cover
    if isinstance(schema, pa.Schema):
        return Schema(schema)
    if ray.__version__ >= "2.5.0":
        if isinstance(schema, rd.Schema):
            if hasattr(schema, "base_schema") and isinstance(
                schema.base_schema, pa.Schema
            ):
                return Schema(schema.base_schema)
            return Schema(list(zip(schema.names, schema.types)))
    raise ValueError(f"{schema} is not supported")


def build_empty(schema: Schema) -> rd.Dataset:
    return rd.from_arrow(_build_empty_arrow(schema))


def add_partition_key(
    df: rd.Dataset, input_schema: Schema, keys: List[str], output_key: str
) -> Tuple[rd.Dataset, Schema]:
    def is_valid_type(tp: pa.DataType) -> bool:
        return (
            pa.types.is_string(tp)
            or pa.types.is_integer(tp)
            or pa.types.is_floating(tp)
            or pa.types.is_date(tp)
            or pa.types.is_time(tp)
            or pa.types.is_timestamp(tp)
            or pa.types.is_boolean(tp)
            or pa.types.is_binary(tp)
        )

    ray_remote_args: Dict[str, Any] = {"num_cpus": 1}

    if len(keys) == 1 and is_valid_type(input_schema[keys[0]].type):

        def add_simple_key(arrow_df: pa.Table) -> pa.Table:  # pragma: no cover
            return arrow_df.append_column(
                output_key,
                arrow_df.column(input_schema.index_of_key(keys[0]))
                .cast(pa.string())
                .fill_null(_RAY_NULL_REPR),
            )

        return df.map_batches(
            add_simple_key, batch_format="pyarrow", **_ZERO_COPY, **ray_remote_args
        ), input_schema + (
            output_key,
            str,
        )
    else:
        key_cols = [input_schema.index_of_key(k) for k in keys]

        def add_key(arrow_df: pa.Table) -> pa.Table:  # pragma: no cover
            fdf = arrow_df.combine_chunks()
            sarr = pa.StructArray.from_arrays(
                [fdf.column(i).combine_chunks() for i in key_cols], keys
            ).tolist()
            sarr = pa.array([pickle.dumps(x) for x in sarr])
            return fdf.append_column(output_key, sarr)

        return df.map_batches(
            add_key, batch_format="pyarrow", **_ZERO_COPY, **ray_remote_args
        ), input_schema + (
            output_key,
            pa.binary(),
        )


def add_coarse_partition_key(
    df: rd.Dataset,
    keys: List[str],
    output_key: str,
    bucket: int,
) -> rd.Dataset:
    ray_remote_args: Dict[str, Any] = {"num_cpus": 1}

    def add_coarse_key(arrow_df: pa.Table) -> pa.Table:  # pragma: no cover
        hdf = arrow_df.select(keys).to_pandas()
        _hash = pd.util.hash_pandas_object(hdf, index=False).mod(bucket)
        return arrow_df.append_column(output_key, pa.Array.from_pandas(_hash))

    return df.map_batches(
        add_coarse_key,
        batch_format="pyarrow",
        **_ZERO_COPY,
        **ray_remote_args,
    )



# flake8: noqa
from .api import *
from .array_dataframe import ArrayDataFrame
from .arrow_dataframe import ArrowDataFrame
from .dataframe import (
    AnyDataFrame,
    DataFrame,
    LocalBoundedDataFrame,
    LocalDataFrame,
    YieldedDataFrame,
)
from .dataframe_iterable_dataframe import (
    IterableArrowDataFrame,
    IterablePandasDataFrame,
    LocalDataFrameIterableDataFrame,
)
from .dataframes import DataFrames
from .function_wrapper import DataFrameFunctionWrapper, fugue_annotated_param
from .iterable_dataframe import IterableDataFrame
from .pandas_dataframe import PandasDataFrame
from .utils import get_column_names, normalize_dataframe_column_names, rename



from typing import Any, Callable

import ibis
import pandas as pd
from ibis.backends.pandas import Backend
from triad.utils.assertion import assert_or_throw

from fugue import (
    DataFrame,
    DataFrames,
    ExecutionEngine,
    NativeExecutionEngine,
    PandasDataFrame,
)
from fugue_ibis._utils import to_ibis_schema, to_schema

from .._compat import IbisTable
from .ibis_engine import IbisEngine, parse_ibis_engine


class PandasIbisEngine(IbisEngine):
    def select(
        self, dfs: DataFrames, ibis_func: Callable[[ibis.BaseBackend], IbisTable]
    ) -> DataFrame:  # pragma: no cover
        pdfs = {k: v.as_pandas() for k, v in dfs.items()}
        be = _BackendWrapper().connect(pdfs)
        be.set_schemas(dfs)
        expr = ibis_func(be)
        schema = to_schema(expr.schema())
        result = expr.execute()
        assert_or_throw(
            isinstance(result, pd.DataFrame), "result must be a pandas DataFrame"
        )
        return PandasDataFrame(result, schema=schema)


@parse_ibis_engine.candidate(
    lambda obj, *args, **kwargs: isinstance(obj, NativeExecutionEngine)
)
def _pd_to_ibis_engine(obj: Any, engine: ExecutionEngine) -> IbisEngine:
    return PandasIbisEngine(engine)


class _BackendWrapper(Backend):
    def set_schemas(self, dfs: DataFrames) -> None:
        self._schemas = {k: to_ibis_schema(v.schema) for k, v in dfs.items()}

    def table(self, name: str, schema: Any = None):
        return super().table(
            name,
            schema=self._schemas[name]
            if schema is None and name in self._schemas
            else schema,
        )



from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import fs as pfs
from dask import dataframe as dd
from fugue._utils.io import FileParser, _get_single_files
from triad.collections.dict import ParamDict
from triad.collections.fs import FileSystem
from triad.collections.schema import Schema
from triad.utils.assertion import assert_or_throw

from fugue_dask.dataframe import DaskDataFrame


def load_df(
    uri: Union[str, List[str]],
    format_hint: Optional[str] = None,
    columns: Any = None,
    fs: Optional[FileSystem] = None,
    **kwargs: Any,
) -> DaskDataFrame:
    if isinstance(uri, str):
        fp = [FileParser(uri, format_hint)]
    else:
        fp = [FileParser(u, format_hint) for u in uri]
    dfs: List[dd.DataFrame] = []
    schema: Any = None
    for f in _get_single_files(fp, fs):
        df, schema = _FORMAT_LOAD[f.file_format](f, columns, **kwargs)
        dfs.append(df)
    return DaskDataFrame(dd.concat(dfs), schema)


def save_df(
    df: DaskDataFrame,
    uri: str,
    format_hint: Optional[str] = None,
    mode: str = "overwrite",
    fs: Optional[FileSystem] = None,
    **kwargs: Any,
) -> None:
    assert_or_throw(
        mode in ["overwrite", "error"],
        lambda: NotImplementedError(f"{mode} is not supported"),
    )
    p = FileParser(uri, format_hint).assert_no_glob()
    if fs is None:
        fs = FileSystem()
    if fs.exists(uri):
        assert_or_throw(mode == "overwrite", FileExistsError(uri))
        try:
            fs.remove(uri)
        except Exception:
            try:
                fs.removetree(uri)
            except Exception:  # pragma: no cover
                pass
    _FORMAT_SAVE[p.file_format](df, p, **kwargs)


def _save_parquet(df: DaskDataFrame, p: FileParser, **kwargs: Any) -> None:
    df.native.to_parquet(p.uri, **{"schema": df.schema.pa_schema, **kwargs})


def _load_parquet(
    p: FileParser, columns: Any = None, **kwargs: Any
) -> Tuple[dd.DataFrame, Any]:
    if columns is None:
        pdf = dd.read_parquet(p.uri, **kwargs)
        schema = Schema(pdf.head(1))
        return pdf, schema
    if isinstance(columns, list):  # column names
        pdf = dd.read_parquet(p.uri, columns=columns, **kwargs)
        schema = Schema(pdf.head(1))
        return pdf, schema
    schema = Schema(columns)
    pdf = dd.read_parquet(p.uri, columns=schema.names, **kwargs)
    return pdf, schema


def _save_csv(df: DaskDataFrame, p: FileParser, **kwargs: Any) -> None:
    df.native.to_csv(
        pfs.path.combine(p.uri, "*.csv"), **{"index": False, "header": False, **kwargs}
    )


def _safe_load_csv(path: str, **kwargs: Any) -> dd.DataFrame:
    try:
        return dd.read_csv(path, **kwargs)
    except (IsADirectoryError, PermissionError):
        return dd.read_csv(pfs.path.combine(path, "*.csv"), **kwargs)


def _load_csv(  # noqa: C901
    p: FileParser, columns: Any = None, **kwargs: Any
) -> Tuple[dd.DataFrame, Any]:
    kw = ParamDict(kwargs)
    infer_schema = kw.get("infer_schema", False)
    if infer_schema and columns is not None and not isinstance(columns, list):
        raise ValueError("can't set columns as a schema when infer schema is true")
    if not infer_schema:
        kw["dtype"] = object
    if "infer_schema" in kw:
        del kw["infer_schema"]
    header: Any = False
    if "header" in kw:
        header = kw["header"]
        del kw["header"]
    if str(header) in ["True", "0"]:
        pdf = _safe_load_csv(p.uri, **{"header": 0, **kw})
        if columns is None:
            return pdf, None
        if isinstance(columns, list):  # column names
            return pdf[columns], None
        schema = Schema(columns)
        return pdf[schema.names], schema
    if header is None or str(header) == "False":
        if columns is None:
            raise ValueError("columns must be set if without header")
        if isinstance(columns, list):  # column names
            pdf = _safe_load_csv(p.uri, **{"header": None, "names": columns, **kw})
            return pdf, None
        schema = Schema(columns)
        pdf = _safe_load_csv(p.uri, **{"header": None, "names": schema.names, **kw})
        return pdf, schema
    else:
        raise NotImplementedError(f"{header} is not supported")


def _save_json(df: DaskDataFrame, p: FileParser, **kwargs: Any) -> None:
    df.native.to_json(pfs.path.combine(p.uri, "*.json"), **kwargs)


def _safe_load_json(path: str, **kwargs: Any) -> dd.DataFrame:
    try:
        return dd.read_json(path, **kwargs)
    except (IsADirectoryError, PermissionError):
        x = dd.read_json(pfs.path.combine(path, "*.json"), **kwargs)
        print(x.compute())
        return x


def _load_json(
    p: FileParser, columns: Any = None, **kwargs: Any
) -> Tuple[dd.DataFrame, Any]:
    pdf = _safe_load_json(p.uri, **kwargs).reset_index(drop=True)
    if columns is None:
        return pdf, None
    if isinstance(columns, list):  # column names
        return pdf[columns], None
    schema = Schema(columns)
    return pdf[schema.names], schema


_FORMAT_LOAD: Dict[str, Callable[..., Tuple[dd.DataFrame, Any]]] = {
    "csv": _load_csv,
    "parquet": _load_parquet,
    "json": _load_json,
}

_FORMAT_SAVE: Dict[str, Callable] = {
    "csv": _save_csv,
    "parquet": _save_parquet,
    "json": _save_json,
}



import copy
from typing import Any, Callable, Dict, List, Optional, no_type_check

from triad import ParamDict, to_uuid
from triad.utils.convert import get_caller_global_local_vars, to_function, to_instance

from fugue._utils.registry import fugue_plugin
from fugue.dataframe import DataFrames
from fugue.dataframe.function_wrapper import DataFrameFunctionWrapper
from fugue.exceptions import FugueInterfacelessError
from fugue.extensions._utils import (
    load_namespace_extensions,
    parse_validation_rules_from_comment,
    to_validation_rules,
)
from fugue.extensions.outputter.outputter import Outputter

_OUTPUTTER_REGISTRY = ParamDict()


@fugue_plugin
def parse_outputter(obj: Any) -> Any:
    """Parse an object to another object that can be converted to a Fugue
    :class:`~fugue.extensions.outputter.outputter.Outputter`.

    .. admonition:: Examples

        .. code-block:: python

            from fugue import Outputter, parse_outputter, FugueWorkflow
            from triad import to_uuid

            class My(Outputter):
                def __init__(self, x):
                    self.x = x

                def process(self, dfs):
                    raise NotImplementedError

                def __uuid__(self) -> str:
                    return to_uuid(super().__uuid__(), self.x)

            @parse_outputter.candidate(
                lambda x: isinstance(x, str) and x.startswith("-*"))
            def _parse(obj):
                return My(obj)

            dag = FugueWorkflow()
            dag.df([[0]], "a:int").output("-*abc")
            # ==  dag.df([[0]], "a:int").output(My("-*abc"))

            dag.run()
    """
    if isinstance(obj, str) and obj in _OUTPUTTER_REGISTRY:
        return _OUTPUTTER_REGISTRY[obj]
    return obj


def register_outputter(alias: str, obj: Any, on_dup: int = ParamDict.OVERWRITE) -> None:
    """Register outputter with an alias.

    :param alias: alias of the processor
    :param obj: the object that can be converted to
        :class:`~fugue.extensions.outputter.outputter.Outputter`
    :param on_dup: see :meth:`triad.collections.dict.ParamDict.update`
        , defaults to ``ParamDict.OVERWRITE``

    .. tip::

        Registering an extension with an alias is particularly useful for projects
        such as libraries. This is because by using alias, users don't have to
        import the specific extension, or provide the full path of the extension.
        It can make user's code less dependent and easy to understand.

    .. admonition:: New Since
        :class: hint

        **0.6.0**

    .. seealso::

        Please read
        :doc:`Outputter Tutorial <tutorial:tutorials/extensions/outputter>`

    .. admonition:: Examples

        Here is an example how you setup your project so your users can
        benefit from this feature. Assume your project name is ``pn``

        The processor implementation in file ``pn/pn/outputters.py``

        .. code-block:: python

            from fugue import DataFrame

            def my_outputter(df:DataFrame) -> None:
                print(df)

        Then in ``pn/pn/__init__.py``

        .. code-block:: python

            from .outputters import my_outputter
            from fugue import register_outputter

            def register_extensions():
                register_outputter("mo", my_outputter)
                # ... register more extensions

            register_extensions()

        In users code:

        .. code-block:: python

            import pn  # register_extensions will be called
            from fugue import FugueWorkflow

            dag = FugueWorkflow()
            # use my_outputter by alias
            dag.df([[0]],"a:int").output("mo")
            dag.run()
    """
    _OUTPUTTER_REGISTRY.update({alias: obj}, on_dup=on_dup)


def outputter(**validation_rules: Any) -> Callable[[Any], "_FuncAsOutputter"]:
    """Decorator for outputters

    Please read
    :doc:`Outputter Tutorial <tutorial:tutorials/extensions/outputter>`
    """

    def deco(func: Callable) -> "_FuncAsOutputter":
        return _FuncAsOutputter.from_func(
            func, validation_rules=to_validation_rules(validation_rules)
        )

    return deco


def _to_outputter(
    obj: Any,
    global_vars: Optional[Dict[str, Any]] = None,
    local_vars: Optional[Dict[str, Any]] = None,
    validation_rules: Optional[Dict[str, Any]] = None,
) -> Outputter:
    global_vars, local_vars = get_caller_global_local_vars(global_vars, local_vars)
    load_namespace_extensions(obj)
    obj = parse_outputter(obj)
    exp: Optional[Exception] = None
    if validation_rules is None:
        validation_rules = {}
    try:
        return copy.copy(
            to_instance(obj, Outputter, global_vars=global_vars, local_vars=local_vars)
        )
    except Exception as e:
        exp = e
    try:
        f = to_function(obj, global_vars=global_vars, local_vars=local_vars)
        # this is for string expression of function with decorator
        if isinstance(f, Outputter):
            return copy.copy(f)
        # this is for functions without decorator
        return _FuncAsOutputter.from_func(f, validation_rules=validation_rules)
    except Exception as e:
        exp = e
    raise FugueInterfacelessError(f"{obj} is not a valid outputter", exp)


class _FuncAsOutputter(Outputter):
    @property
    def validation_rules(self) -> Dict[str, Any]:
        return self._validation_rules  # type: ignore

    @no_type_check
    def process(self, dfs: DataFrames) -> None:
        args: List[Any] = []
        kwargs: Dict[str, Any] = {}
        if self._engine_param is not None:
            args.append(self._engine_param.to_input(self.execution_engine))
        if self._use_dfs:
            args.append(dfs)
        else:
            if not dfs.has_key:
                args += dfs.values()
            else:
                kwargs.update(dfs)
        kwargs.update(self.params)
        return self._wrapper.run(args=args, kwargs=kwargs, ctx=self.execution_engine)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self._wrapper(*args, **kwargs)  # type: ignore

    @no_type_check
    def __uuid__(self) -> str:
        return to_uuid(self._wrapper, self._engine_param, self._use_dfs)

    @no_type_check
    @staticmethod
    def from_func(
        func: Callable, validation_rules: Dict[str, Any]
    ) -> "_FuncAsOutputter":
        validation_rules.update(parse_validation_rules_from_comment(func))
        tr = _FuncAsOutputter()
        tr._wrapper = DataFrameFunctionWrapper(  # type: ignore
            func, "^e?(c|[dlspq]+)x*z?$", "^n$"
        )
        tr._engine_param = (
            tr._wrapper._params.get_value_by_index(0)
            if tr._wrapper.input_code.startswith("e")
            else None
        )
        tr._use_dfs = "c" in tr._wrapper.input_code
        tr._validation_rules = validation_rules
        return tr



from types import GeneratorType
from typing import Any, Iterable, List

from ..exceptions import FugueDatasetEmptyError
from .bag import LocalBoundedBag


class ArrayBag(LocalBoundedBag):
    def __init__(self, data: Any, copy: bool = True):
        if isinstance(data, list):
            self._native = list(data) if copy else data
        elif isinstance(data, (GeneratorType, Iterable)):
            self._native = list(data)
        else:
            raise ValueError(f"{type(data)} can't be converted to ArrayBag")
        super().__init__()

    @property
    def native(self) -> List[Any]:
        """The underlying Python list object"""
        return self._native

    @property
    def empty(self) -> bool:
        return len(self._native) == 0

    def count(self) -> int:
        return len(self._native)

    def peek(self) -> Any:
        if self.count() == 0:
            raise FugueDatasetEmptyError()
        return self._native[0]

    def as_array(self) -> List[Any]:
        return list(self._native)

    def head(self, n: int) -> LocalBoundedBag:
        return ArrayBag(self._native[:n])



import os
from typing import Any, Iterable, List, Optional, Union

from duckdb import DuckDBPyConnection
from triad import ParamDict, Schema
from triad.collections.fs import FileSystem
from triad.utils.assertion import assert_or_throw

from fugue._utils.io import FileParser, load_df, save_df
from fugue.collections.sql import TempTableName
from fugue.dataframe import ArrowDataFrame, LocalBoundedDataFrame
from fugue_duckdb._utils import (
    encode_value_to_expr,
    to_duck_type,
    encode_column_name,
    encode_column_names,
)
from fugue_duckdb.dataframe import DuckDataFrame


def _get_single_files(
    fp: Iterable[FileParser], fs: FileSystem, fmt: str
) -> Iterable[FileParser]:
    def _isdir(d: str) -> bool:
        try:
            return fs.isdir(d)
        except Exception:  # pragma: no cover
            return False

    for f in fp:
        if f.glob_pattern == "" and _isdir(f.uri):
            yield f.with_glob("*." + fmt, fmt)
        else:
            yield f


class DuckDBIO:
    def __init__(self, fs: FileSystem, con: DuckDBPyConnection) -> None:
        self._con = con
        self._fs = fs
        self._format_load = {"csv": self._load_csv, "parquet": self._load_parquet}
        self._format_save = {"csv": self._save_csv, "parquet": self._save_parquet}

    def load_df(
        self,
        uri: Union[str, List[str]],
        format_hint: Optional[str] = None,
        columns: Any = None,
        **kwargs: Any,
    ) -> LocalBoundedDataFrame:
        for k in kwargs.keys():
            assert_or_throw(k.isidentifier(), ValueError(f"{k} is invalid"))
        if isinstance(uri, str):
            fp = [FileParser(uri, format_hint)]
        else:
            fp = [FileParser(u, format_hint) for u in uri]
        if fp[0].file_format not in self._format_load:
            return load_df(
                uri, format_hint=format_hint, columns=columns, fs=self._fs, **kwargs
            )
        dfs: List[DuckDataFrame] = []
        for f in _get_single_files(fp, self._fs, fp[0].file_format):
            df = self._format_load[f.file_format](f, columns, **kwargs)
            dfs.append(df)
        rel = dfs[0].native
        for i in range(1, len(dfs)):
            rel = rel.union(dfs[i].native)
        return DuckDataFrame(rel)

    def save_df(
        self,
        df: DuckDataFrame,
        uri: str,
        format_hint: Optional[str] = None,
        mode: str = "overwrite",
        **kwargs: Any,
    ) -> None:
        for k in kwargs.keys():
            assert_or_throw(k.isidentifier(), ValueError(f"{k} is invalid"))
        assert_or_throw(
            mode in ["overwrite", "error"],
            NotImplementedError(f"{mode} is not supported"),
        )
        p = FileParser(uri, format_hint).assert_no_glob()
        if (p.file_format not in self._format_save) or ("partition_cols" in kwargs):
            self._fs.makedirs(os.path.dirname(uri), recreate=True)
            ldf = ArrowDataFrame(df.as_arrow())
            return save_df(
                ldf, uri=uri, format_hint=format_hint, mode=mode, fs=self._fs, **kwargs
            )
        fs = self._fs
        if fs.exists(uri):
            assert_or_throw(mode == "overwrite", FileExistsError(uri))
            try:
                fs.remove(uri)
            except Exception:
                try:
                    fs.removetree(uri)
                except Exception:  # pragma: no cover
                    pass
        if not fs.exists(p.parent):
            fs.makedirs(p.parent, recreate=True)
        self._format_save[p.file_format](df, p, **kwargs)

    def _save_csv(self, df: DuckDataFrame, p: FileParser, **kwargs: Any):
        dn = TempTableName()
        df.native.create_view(dn.key)
        kw = ParamDict({k.lower(): v for k, v in kwargs.items()})
        kw["header"] = 1 if kw.pop("header", False) else 0
        params: List[str] = []
        for k, v in kw.items():
            params.append(f"{k.upper()} " + encode_value_to_expr(v))
        pm = ", ".join(params)
        query = f"COPY {dn.key} TO {encode_value_to_expr(p.uri)} WITH ({pm})"
        self._con.execute(query)

    def _load_csv(  # noqa: C901
        self, p: FileParser, columns: Any = None, **kwargs: Any
    ) -> DuckDataFrame:
        kw = ParamDict({k.lower(): v for k, v in kwargs.items()})
        infer_schema = kw.pop("infer_schema", False)
        header = kw.pop("header", False)
        assert_or_throw(
            not (columns is None and not header),
            ValueError("when csv has no header, columns must be specified"),
        )
        kw.pop("auto_detect", None)
        params: List[str] = [encode_value_to_expr(p.uri_with_glob)]
        kw["header"] = 1 if header else 0
        kw["auto_detect"] = 1 if infer_schema else 0
        if infer_schema:
            for k, v in kw.items():
                params.append(f"{k}=" + encode_value_to_expr(v))
            pm = ", ".join(params)
            if header:
                if columns is None:
                    cols = "*"
                elif isinstance(columns, list):
                    cols = ", ".join(encode_column_names(columns))
                else:
                    raise ValueError(
                        "columns can't be schema when infer_schema is true"
                    )
                query = f"SELECT {cols} FROM read_csv_auto({pm})"
                return DuckDataFrame(self._con.from_query(query))
            else:
                if isinstance(columns, list):
                    pass
                else:
                    raise ValueError(
                        "columns can't be schema when infer_schema is true"
                    )
                query = f"SELECT * FROM read_csv_auto({pm})"
                tdf = DuckDataFrame(self._con.from_query(query))
                rn = dict(zip(tdf.columns, columns))
                return tdf.rename(rn)  # type: ignore
        else:
            if header:
                kw["ALL_VARCHAR"] = 1
                if columns is None:
                    cols = "*"
                elif isinstance(columns, list):
                    cols = ", ".join(encode_column_names(columns))
                else:
                    cols = "*"
                for k, v in kw.items():
                    params.append(f"{k}=" + encode_value_to_expr(v))
                pm = ", ".join(params)
                query = f"SELECT {cols} FROM read_csv_auto({pm})"
                res = DuckDataFrame(self._con.from_query(query))
                if isinstance(columns, list):
                    res = res[columns]  # type: ignore
                elif columns is not None:
                    res = res[Schema(columns).names].alter_columns(  # type: ignore
                        columns
                    )
                return res
            else:
                if isinstance(columns, list):
                    schema = Schema([(x, str) for x in columns])
                else:
                    schema = Schema(columns)
                kw["columns"] = {f.name: to_duck_type(f.type) for f in schema.fields}
                for k, v in kw.items():
                    params.append(encode_column_name(k) + "=" + encode_value_to_expr(v))
                pm = ", ".join(params)
                query = f"SELECT * FROM read_csv({pm})"
                return DuckDataFrame(self._con.from_query(query))

    def _save_parquet(self, df: DuckDataFrame, p: FileParser, **kwargs: Any):
        dn = TempTableName()
        df.native.create_view(dn.key)
        kw = ParamDict({k.lower(): v for k, v in kwargs.items()})
        kw["format"] = "parquet"
        params: List[str] = []
        for k, v in kw.items():
            params.append(f"{k.upper()} " + encode_value_to_expr(v))
        pm = ", ".join(params)
        query = f"COPY {dn.key} TO {encode_value_to_expr(p.uri)}"
        if len(params) > 0:
            query += f" WITH ({pm})"
        self._con.execute(query)

    def _load_parquet(
        self, p: FileParser, columns: Any = None, **kwargs: Any
    ) -> DuckDataFrame:
        kw = ParamDict({k.lower(): v for k, v in kwargs.items()})
        params: List[str] = [encode_value_to_expr(p.uri_with_glob)]
        if isinstance(columns, list):
            cols = ", ".join(encode_column_names(columns))
        else:
            cols = "*"
        assert_or_throw(
            len(kw) == 0,
            NotImplementedError("can't take extra parameters for loading parquet"),
        )
        # for k, v in kw.items():
        #    params.append(f"{k}=" + encode_value_to_expr(v))
        pm = ", ".join(params)
        query = f"SELECT {cols} FROM parquet_scan([{pm}])"
        res = DuckDataFrame(self._con.from_query(query))
        return (
            res  # type: ignore
            if isinstance(columns, list) or columns is None
            else res.alter_columns(columns)
        )



from typing import Any, Dict, Union

from fugue.collections.partition import PartitionCursor, PartitionSpec
from fugue.dataframe import DataFrame, DataFrames
from fugue.execution.execution_engine import ExecutionEngine
from fugue.extensions._utils import validate_input_schema, validate_partition_spec
from fugue.rpc import RPCClient, RPCServer
from triad.collections import ParamDict, Schema
from triad.utils.convert import get_full_type_path
from triad.utils.hash import to_uuid


class ExtensionContext(object):
    # pylint: disable=E1101
    """Context variables that extensions can access. It's also the base
    class of all extensions.
    """

    @property
    def params(self) -> ParamDict:
        """Parameters set for using this extension.

        .. admonition:: Examples

            >>> FugueWorkflow().df(...).transform(using=dummy, params={"a": 1})

        You will get ``{"a": 1}`` as `params` in the ``dummy`` transformer
        """
        return self._params  # type: ignore

    @property
    def workflow_conf(self) -> ParamDict:
        """Workflow level configs, this is accessible even in
        :class:`~fugue.extensions.transformer.transformer.Transformer` and
        :class:`~fugue.extensions.transformer.transformer.CoTransformer`

        .. admonition:: Examples

            >>> dag = FugueWorkflow().df(...).transform(using=dummy)
            >>> dag.run(NativeExecutionEngine(conf={"b": 10}))

        You will get ``{"b": 10}`` as `workflow_conf` in the ``dummy`` transformer
        on both driver and workers.
        """
        if "_workflow_conf" in self.__dict__:
            return self._workflow_conf  # type: ignore
        return self.execution_engine.conf

    @property
    def execution_engine(self) -> ExecutionEngine:
        """Execution engine for the current execution, this is only available on
        driver side
        """
        return self._execution_engine  # type: ignore

    @property
    def output_schema(self) -> Schema:
        """Output schema of the operation. This is accessible for all extensions (
        if defined), and on both driver and workers
        """
        return self._output_schema  # type: ignore

    @property
    def key_schema(self) -> Schema:
        """Partition keys schema, this is for transformers only, and available on both
        driver and workers
        """
        return self._key_schema  # type: ignore

    @property
    def partition_spec(self) -> PartitionSpec:
        """Partition specification, this is for all extensions except for creators,
        and available on both driver and workers
        """
        return self._partition_spec  # type: ignore

    @property
    def cursor(self) -> PartitionCursor:
        """Cursor of the current logical partition, this is for transformers only,
        and only available on worker side
        """
        return self._cursor  # type: ignore

    @property
    def has_callback(self) -> bool:
        """Whether this transformer has callback"""
        return (
            "_has_rpc_client" in self.__dict__ and self._has_rpc_client  # type: ignore
        )

    @property
    def callback(self) -> RPCClient:
        """RPC client to talk to driver, this is for transformers only,
        and available on both driver and workers
        """
        return self._rpc_client  # type: ignore

    @property
    def rpc_server(self) -> RPCServer:
        """RPC client to talk to driver, this is for transformers only,
        and available on both driver and workers
        """
        return self._rpc_server  # type: ignore

    @property
    def validation_rules(self) -> Dict[str, Any]:
        """Extension input validation rules defined by user"""
        return {}

    def validate_on_compile(self) -> None:
        validate_partition_spec(self.partition_spec, self.validation_rules)

    def validate_on_runtime(self, data: Union[DataFrame, DataFrames]) -> None:
        if isinstance(data, DataFrame):
            validate_input_schema(data.schema, self.validation_rules)
        else:
            for df in data.values():
                validate_input_schema(df.schema, self.validation_rules)

    def __uuid__(self) -> str:
        return to_uuid(get_full_type_path(self))



from typing import Any, Callable, Dict, List, Optional, Union

import pyspark.sql as ps
from pyspark.sql import SparkSession
from triad.collections import Schema
from triad.collections.dict import ParamDict
from triad.collections.fs import FileSystem
from triad.utils.assertion import assert_or_throw

from fugue._utils.io import FileParser, save_df
from fugue.collections.partition import PartitionSpec
from fugue.dataframe import DataFrame
from fugue_spark.dataframe import SparkDataFrame

from .convert import to_schema, to_spark_schema


class SparkIO(object):
    def __init__(self, spark_session: SparkSession, fs: FileSystem):
        self._session = spark_session
        self._fs = fs
        self._loads: Dict[str, Callable[..., DataFrame]] = {
            "csv": self._load_csv,
            "parquet": self._load_parquet,
            "json": self._load_json,
        }

    def load_df(
        self,
        uri: Union[str, List[str]],
        format_hint: Optional[str] = None,
        columns: Any = None,
        **kwargs: Any,
    ) -> DataFrame:
        if isinstance(uri, str):
            fp = [FileParser(uri, format_hint)]
        else:
            fp = [FileParser(u, format_hint) for u in uri]
        fmts = list(set(f.file_format for f in fp))  # noqa: C401
        assert_or_throw(
            len(fmts) == 1, NotImplementedError("can't support multiple formats")
        )
        fmt = fmts[0]
        files = [f.uri for f in fp]
        return self._loads[fmt](files, columns, **kwargs)

    def save_df(
        self,
        df: SparkDataFrame,
        uri: str,
        format_hint: Optional[str] = None,
        partition_spec: Optional[PartitionSpec] = None,
        mode: str = "overwrite",
        force_single: bool = False,
        **kwargs: Any,
    ) -> None:
        partition_spec = partition_spec or PartitionSpec()
        if not force_single:
            p = FileParser(uri, format_hint)
            writer = self._get_writer(df.native, partition_spec)
            writer.format(p.file_format).options(**kwargs).mode(mode)
            writer.save(uri)
        else:
            ldf = df.as_local()
            save_df(ldf, uri, format_hint=format_hint, mode=mode, fs=self._fs, **kwargs)

    def _get_writer(
        self, sdf: ps.DataFrame, partition_spec: PartitionSpec
    ) -> ps.DataFrameWriter:
        # will not use bucket by because here we don't save as table
        num = int(partition_spec.num_partitions)  # TODO: this is hacky
        by = partition_spec.partition_by
        if num == 0 and len(by) == 0:
            return sdf.write
        elif num > 0 and len(by) == 0:
            return sdf.repartition(num).write
        elif num == 0 and len(by) > 0:
            return sdf.write.partitionBy(*by)
        return sdf.repartition(num, *by).write.partitionBy(*by)

    def _load_parquet(
        self, p: List[str], columns: Any = None, **kwargs: Any
    ) -> DataFrame:
        sdf = self._session.read.parquet(*p, **kwargs)
        if columns is None:
            return SparkDataFrame(sdf)
        if isinstance(columns, list):  # column names
            return SparkDataFrame(sdf)[columns]
        schema = Schema(columns)
        return SparkDataFrame(sdf[schema.names], schema)

    def _load_csv(  # noqa: C901
        self, p: List[str], columns: Any = None, **kwargs: Any
    ) -> DataFrame:
        kw = ParamDict(kwargs)
        infer_schema = kw.get("infer_schema", False)
        if infer_schema and columns is not None and not isinstance(columns, list):
            raise ValueError("can't set columns as a schema when infer schema is true")
        if infer_schema:
            kw["inferSchema"] = True
        if "infer_schema" in kw:
            del kw["infer_schema"]
        header = str(kw.get_or_none("header", object)).lower()
        if "header" in kw:
            del kw["header"]
        reader = self._session.read.format("csv")
        reader.options(**kw)
        if header == "true":
            reader.option("header", "true")
            if columns is None:
                return SparkDataFrame(reader.load(p))
            if isinstance(columns, list):  # column names
                return SparkDataFrame(reader.load(p)[columns])
            schema = Schema(columns)
            return SparkDataFrame(reader.load(p)[schema.names], schema)
        if header in ["false", "none"]:
            reader.option("header", "false")
            if columns is None:
                raise ValueError("columns must be set if without header")
            if isinstance(columns, list):  # column names
                sdf = reader.load(p)
                inferred = to_schema(sdf)
                renames = [f"{k} AS {v}" for k, v in zip(inferred.names, columns)]
                return SparkDataFrame(sdf.selectExpr(*renames))
            schema = Schema(columns)
            sdf = reader.schema(to_spark_schema(schema)).load(p)
            return SparkDataFrame(sdf, schema)
        else:
            raise NotImplementedError(f"{header} is not supported")

    def _load_json(self, p: List[str], columns: Any = None, **kwargs: Any) -> DataFrame:
        reader = self._session.read.format("json")
        reader.options(**kwargs)
        if columns is None:
            return SparkDataFrame(reader.load(p))
        if isinstance(columns, list):  # column names
            return SparkDataFrame(reader.load(p))[columns]
        schema = Schema(columns)
        return SparkDataFrame(reader.load(p)[schema.names], schema)



import base64
import json
import os
import pickle
from typing import Any, Iterable, List, Optional, Tuple

import pandas as pd
import pyarrow as pa
from fs import open_fs
from triad import FileSystem, Schema
from triad.collections.schema import SchemaError
from triad.exceptions import InvalidOperationError
from triad.utils.assertion import assert_arg_not_none
from triad.utils.assertion import assert_or_throw as aot

from .api import get_column_names, normalize_column_names, rename, as_fugue_df
from .array_dataframe import ArrayDataFrame
from .dataframe import DataFrame, LocalBoundedDataFrame
from .pandas_dataframe import PandasDataFrame

# For backward compatibility, TODO: remove!
get_dataframe_column_names = get_column_names
normalize_dataframe_column_names = normalize_column_names
rename_dataframe_column_names = rename


def _pa_type_eq(t1: pa.DataType, t2: pa.DataType) -> bool:
    # should ignore the name difference of list
    # e.g. list<item: string> == list<l: string>
    if pa.types.is_list(t1) and pa.types.is_list(t2):  # pragma: no cover
        return _pa_type_eq(t1.value_type, t2.value_type)
    return t1 == t2


def _schema_eq(s1: Schema, s2: Schema) -> bool:
    if s1 == s2:
        return True
    return s1.names == s2.names and all(
        _pa_type_eq(f1.type, f2.type) for f1, f2 in zip(s1.fields, s2.fields)
    )


def _df_eq(
    df: DataFrame,
    data: Any,
    schema: Any = None,
    digits=8,
    check_order: bool = False,
    check_schema: bool = True,
    check_content: bool = True,
    no_pandas: bool = False,
    throw=False,
) -> bool:
    """Compare if two dataframes are equal. Is for internal, unit test
    purpose only. It will convert both dataframes to
    :class:`~fugue.dataframe.dataframe.LocalBoundedDataFrame`, so it assumes
    both dataframes are small and fast enough to convert. DO NOT use it
    on critical or expensive tasks.

    :param df: first data frame
    :param data: :ref:`DataFrame like
      <tutorial:tutorials/advanced/x-like:dataframe>` object
    :param schema: :ref:`Schema like
      <tutorial:tutorials/advanced/x-like:schema>` object, defaults to None
    :param digits: precision on float number comparison, defaults to 8
    :param check_order: if to compare the row orders, defaults to False
    :param check_schema: if compare schemas, defaults to True
    :param check_content: if to compare the row values, defaults to True
    :param no_pandas: if true, it will compare the string representations of the
      dataframes, otherwise, it will convert both to pandas dataframe to compare,
      defaults to False
    :param throw: if to throw error if not equal, defaults to False
    :return: if they equal
    """
    df1 = as_fugue_df(df).as_local_bounded()
    if schema is not None:
        df2 = as_fugue_df(data, schema=schema).as_local_bounded()
    else:
        df2 = as_fugue_df(data).as_local_bounded()
    try:
        assert (
            df1.count() == df2.count()
        ), f"count mismatch {df1.count()}, {df2.count()}"
        assert not check_schema or _schema_eq(
            df.schema, df2.schema
        ), f"schema mismatch {df.schema.pa_schema}, {df2.schema.pa_schema}"
        if not check_content:
            return True
        if no_pandas:
            dd1 = [[x.__repr__()] for x in df1.as_array_iterable(type_safe=True)]
            dd2 = [[x.__repr__()] for x in df2.as_array_iterable(type_safe=True)]
            d1 = pd.DataFrame(dd1, columns=["data"])
            d2 = pd.DataFrame(dd2, columns=["data"])
        else:
            d1 = df1.as_pandas()
            d2 = df2.as_pandas()
        if not check_order:
            d1 = d1.sort_values(df1.columns)
            d2 = d2.sort_values(df1.columns)
        d1 = d1.reset_index(drop=True)
        d2 = d2.reset_index(drop=True)
        pd.testing.assert_frame_equal(
            d1, d2, rtol=0, atol=10 ** (-digits), check_dtype=False, check_exact=False
        )
        return True
    except AssertionError:
        if throw:
            raise
        return False


def pickle_df(df: DataFrame) -> bytes:
    """Pickles a dataframe to bytes array. It firstly converts the dataframe
    local bounded, and then serialize the underlying data.

    :param df: input DataFrame
    :return: pickled binary data

    .. note::

        Be careful to use on large dataframes or non-local, un-materialized dataframes,
        it can be slow. You should always use :func:`.unpickle_df` to deserialize.
    """
    df = df.as_local_bounded()
    o: List[Any] = [df.schema]
    if isinstance(df, PandasDataFrame):
        o.append("p")
        o.append(df.native)
    else:
        o.append("a")
        o.append(df.as_array())
    return pickle.dumps(o)


def serialize_df(
    df: Optional[DataFrame],
    threshold: int = -1,
    file_path: Optional[str] = None,
    fs: Optional[FileSystem] = None,
) -> str:
    """Serialize input dataframe to base64 string or to file
    if it's larger than threshold

    :param df: input DataFrame
    :param threshold: file byte size threshold, defaults to -1
    :param file_path: file path to store the data (used only if the serialized data
      is larger than ``threshold``), defaults to None
    :param fs: :class:`~triad:triad.collections.fs.FileSystem`, defaults to None
    :raises InvalidOperationError: if file is large but ``file_path`` is not provided
    :return: a json string either containing the base64 data or the file path

    .. note::

        If fs is not provided but it needs to write to disk, then it will use
        :meth:`~fs:fs.opener.registry.Registry.open_fs` to try to open the file to
        write.
    """
    if df is None:
        return json.dumps(dict())
    data = pickle_df(df)
    size = len(data)
    if threshold < 0 or size <= threshold:
        res = dict(data=base64.b64encode(data).decode())
    else:
        if file_path is None:
            raise InvalidOperationError("file_path is not provided")
        if fs is None:
            with open_fs(
                os.path.dirname(file_path), writeable=True, create=False
            ) as _fs:
                _fs.writebytes(os.path.basename(file_path), data)
        else:
            fs.writebytes(file_path, data)
        res = dict(path=file_path)
    return json.dumps(res)


def unpickle_df(stream: bytes) -> LocalBoundedDataFrame:
    """Unpickles a dataframe from bytes array.

    :param stream: binary data
    :return: unpickled dataframe

    .. note::

        The data must be serialized by :func:`.pickle_df` to deserialize.
    """
    o = pickle.loads(stream)
    schema = o[0]
    if o[1] == "p":
        return PandasDataFrame(o[2], schema)
    if o[1] == "a":
        return ArrayDataFrame(o[2], schema)
    raise NotImplementedError(  # pragma: no cover
        f"{o[1]} is not supported for unpickle"
    )


def deserialize_df(
    json_str: str, fs: Optional[FileSystem] = None
) -> Optional[LocalBoundedDataFrame]:
    """Deserialize json string to
    :class:`~fugue.dataframe.dataframe.LocalBoundedDataFrame`

    :param json_str: json string containing the base64 data or a file path
    :param fs: :class:`~triad:triad.collections.fs.FileSystem`, defaults to None
    :raises ValueError: if the json string is invalid, not generated from
      :func:`~.serialize_df`
    :return: :class:`~fugue.dataframe.dataframe.LocalBoundedDataFrame` if ``json_str``
      contains a dataframe or None if its valid but contains no data
    """
    d = json.loads(json_str)
    if len(d) == 0:
        return None
    if "data" in d:
        return unpickle_df(base64.b64decode(d["data"].encode()))
    elif "path" in d:
        if fs is None:
            with open_fs(os.path.dirname(d["path"]), create=False) as _fs:
                return unpickle_df(_fs.readbytes(os.path.basename(d["path"])))
        return unpickle_df(fs.readbytes(d["path"]))
    raise ValueError(f"{json_str} is invalid")


def get_join_schemas(
    df1: DataFrame, df2: DataFrame, how: str, on: Optional[Iterable[str]]
) -> Tuple[Schema, Schema]:
    """Get :class:`~triad:triad.collections.schema.Schema` object after
    joining ``df1`` and ``df2``. If ``on`` is not empty, it's mainly for
    validation purpose.

    :param df1: first dataframe
    :param df2: second dataframe
    :param how: can accept ``semi``, ``left_semi``, ``anti``, ``left_anti``,
      ``inner``, ``left_outer``, ``right_outer``, ``full_outer``, ``cross``
    :param on: it can always be inferred, but if you provide, it will be
      validated agained the inferred keys.
    :return: the pair key schema and schema after join

    .. note::

        In Fugue, joined schema can always be inferred because it always uses the
        input dataframes' common keys as the join keys. So you must make sure to
        :meth:`~fugue.dataframe.dataframe.DataFrame.rename` to input dataframes so
        they follow this rule.
    """
    assert_arg_not_none(how, "how")
    how = how.lower()
    aot(
        how
        in [
            "semi",
            "left_semi",
            "anti",
            "left_anti",
            "inner",
            "left_outer",
            "right_outer",
            "full_outer",
            "cross",
        ],
        ValueError(f"{how} is not a valid join type"),
    )
    on = list(on) if on is not None else []
    aot(len(on) == len(set(on)), f"{on} has duplication")
    if how != "cross" and len(on) == 0:
        other = set(df2.columns)
        on = [c for c in df1.columns if c in other]
        aot(
            len(on) > 0,
            lambda: SchemaError(
                f"no common columns between {df1.columns} and {df2.columns}"
            ),
        )
    schema2 = df2.schema
    aot(
        how != "outer",
        ValueError(
            "'how' must use left_outer, right_outer, full_outer for outer joins"
        ),
    )
    if how in ["semi", "left_semi", "anti", "left_anti"]:
        schema2 = schema2.extract(on)
    aot(
        on in df1.schema and on in schema2,
        lambda: SchemaError(
            f"{on} is not the intersection of {df1.schema} & {df2.schema}"
        ),
    )
    cm = df1.schema.intersect(on)
    if how == "cross":
        cs = df1.schema.intersect(schema2)
        aot(
            len(cs) == 0,
            SchemaError(f"invalid cross join, two dataframes have common columns {cs}"),
        )
    else:
        aot(len(on) > 0, SchemaError("join on columns must be specified"))
    return cm, (df1.schema.union(schema2))



import html
from abc import ABC, abstractmethod
from typing import Any, Optional, TypeVar

from triad import ParamDict, SerializableRLock, assert_or_throw

from .._utils.registry import fugue_plugin
from ..exceptions import FugueDatasetEmptyError


AnyDataset = TypeVar("AnyDataset", "Dataset", object)


class Dataset(ABC):
    """The base class of Fugue :class:`~.fugue.dataframe.dataframe.DataFrame`
    and :class:`~.fugue.bag.bag.Bag`.

    .. note::

        This is for internal use only.
    """

    def __init__(self):
        self._metadata: Optional[ParamDict] = None

    @property
    def metadata(self) -> ParamDict:
        """Metadata of the dataset"""
        if self._metadata is None:
            self._metadata = ParamDict()
        return self._metadata

    @property
    def has_metadata(self) -> bool:
        """Whether this dataframe contains any metadata"""
        return self._metadata is not None and len(self._metadata) > 0

    def reset_metadata(self, metadata: Any) -> None:
        """Reset metadata"""
        self._metadata = ParamDict(metadata) if metadata is not None else None

    @property
    @abstractmethod
    def native(self) -> Any:  # pragma: no cover
        """The native object this Dataset class wraps"""
        raise NotImplementedError

    @property
    @abstractmethod
    def is_local(self) -> bool:  # pragma: no cover
        """Whether this dataframe is a local Dataset"""
        raise NotImplementedError

    @property
    @abstractmethod
    def is_bounded(self) -> bool:  # pragma: no cover
        """Whether this dataframe is bounded"""
        raise NotImplementedError

    @property
    @abstractmethod
    def num_partitions(self) -> int:  # pragma: no cover
        """Number of physical partitions of this dataframe.
        Please read |PartitionTutorial|
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def empty(self) -> bool:  # pragma: no cover
        """Whether this dataframe is empty"""
        raise NotImplementedError

    @abstractmethod
    def count(self) -> int:  # pragma: no cover
        """Get number of rows of this dataframe"""
        raise NotImplementedError

    def assert_not_empty(self) -> None:
        """Assert this dataframe is not empty

        :raises FugueDatasetEmptyError: if it is empty
        """
        assert_or_throw(not self.empty, FugueDatasetEmptyError("dataframe is empty"))

    def show(
        self, n: int = 10, with_count: bool = False, title: Optional[str] = None
    ) -> None:
        """Display the Dataset

        :param n: number of rows to print, defaults to 10
        :param with_count: whether to show dataset count, defaults to False
        :param title: title of the dataset, defaults to None

        .. note::

            When ``with_count`` is True, it can trigger expensive calculation for
            a distributed dataframe. So if you call this function directly, you may
            need to :func:`fugue.execution.execution_engine.ExecutionEngine.persist`
            the dataset.
        """
        return get_dataset_display(self).show(n=n, with_count=with_count, title=title)

    def __repr__(self):
        """String representation of the Dataset"""
        return get_dataset_display(self).repr()

    def _repr_html_(self):
        """HTML representation of the Dataset"""
        return get_dataset_display(self).repr_html()


class DatasetDisplay(ABC):
    """The base class for display handlers of :class:`~.Dataset`

    :param ds: the Dataset
    """

    _SHOW_LOCK = SerializableRLock()

    def __init__(self, ds: Dataset):
        self._ds = ds

    @abstractmethod
    def show(
        self, n: int = 10, with_count: bool = False, title: Optional[str] = None
    ) -> None:  # pragma: no cover
        """Show the :class:`~.Dataset`

        :param n: top n items to display, defaults to 10
        :param with_count: whether to display the total count, defaults to False
        :param title: title to display, defaults to None
        """
        raise NotImplementedError

    def repr(self) -> str:
        """The string representation of the :class:`~.Dataset`

        :return: the string representation
        """
        return str(type(self._ds).__name__)

    def repr_html(self) -> str:
        """The HTML representation of the :class:`~.Dataset`

        :return: the HTML representation
        """
        return html.escape(self.repr())


@fugue_plugin
def get_dataset_display(ds: "Dataset") -> DatasetDisplay:  # pragma: no cover
    """Get the display class to display a :class:`~.Dataset`

    :param ds: the Dataset to be displayed
    """

    raise NotImplementedError(f"no matching DatasetDisplay registered for {type(ds)}")



import logging
from typing import Any, Dict, Iterable, List, Optional, Union

import duckdb
from duckdb import DuckDBPyConnection, DuckDBPyRelation
from triad import SerializableRLock
from triad.collections.fs import FileSystem
from triad.utils.assertion import assert_or_throw
from triad.utils.schema import quote_name

from fugue import (
    ArrowDataFrame,
    ExecutionEngine,
    MapEngine,
    NativeExecutionEngine,
    PandasMapEngine,
    SQLEngine,
)
from fugue.collections.partition import PartitionSpec, parse_presort_exp
from fugue.collections.sql import StructuredRawSQL, TempTableName
from fugue.dataframe import DataFrame, DataFrames, LocalBoundedDataFrame
from fugue.dataframe.utils import get_join_schemas

from ._io import DuckDBIO
from ._utils import (
    encode_column_name,
    encode_column_names,
    encode_schema_names,
    encode_value_to_expr,
)
from .dataframe import DuckDataFrame, _duck_as_arrow

_FUGUE_DUCKDB_PRAGMA_CONFIG_PREFIX = "fugue.duckdb.pragma."
_FUGUE_DUCKDB_EXTENSIONS = "fugue.duckdb.extensions"


class DuckDBEngine(SQLEngine):
    """DuckDB SQL backend implementation.

    :param execution_engine: the execution engine this sql engine will run on
    """

    @property
    def dialect(self) -> Optional[str]:
        return "duckdb"

    def select(self, dfs: DataFrames, statement: StructuredRawSQL) -> DataFrame:
        if isinstance(self.execution_engine, DuckExecutionEngine):
            return self._duck_select(dfs, statement)
        else:
            _dfs, _sql = self.encode(dfs, statement)
            return self._other_select(_dfs, _sql)

    def table_exists(self, table: str) -> bool:
        return self._get_table(table) is not None

    def save_table(
        self,
        df: DataFrame,
        table: str,
        mode: str = "overwrite",
        partition_spec: Optional[PartitionSpec] = None,
        **kwargs: Any,
    ) -> None:
        if isinstance(self.execution_engine, DuckExecutionEngine):
            con = self.execution_engine.connection
            tdf: DuckDataFrame = _to_duck_df(
                self.execution_engine, df, create_view=False  # type: ignore
            )
            et = self._get_table(table)
            if et is not None:
                if mode == "overwrite":
                    tp = "VIEW" if et["table_type"] == "VIEW" else "TABLE"
                    tn = encode_column_name(et["table_name"])
                    con.query(f"DROP {tp} {tn}")
                else:
                    raise Exception(f"{table} exists")
            tdf.native.create(table)
        else:  # pragma: no cover
            raise NotImplementedError(
                "save_table can only be used with DuckExecutionEngine"
            )

    def load_table(self, table: str, **kwargs: Any) -> DataFrame:
        if isinstance(self.execution_engine, DuckExecutionEngine):
            return DuckDataFrame(self.execution_engine.connection.table(table))
        else:  # pragma: no cover
            raise NotImplementedError(
                "load_table can only be used with DuckExecutionEngine"
            )

    @property
    def is_distributed(self) -> bool:
        return False

    def _duck_select(self, dfs: DataFrames, statement: StructuredRawSQL) -> DataFrame:
        name_map: Dict[str, str] = {}
        for k, v in dfs.items():
            tdf: DuckDataFrame = _to_duck_df(
                self.execution_engine, v, create_view=True  # type: ignore
            )
            name_map[k] = tdf.alias
        query = statement.construct(name_map, dialect=self.dialect, log=self.log)
        result = self.execution_engine.connection.query(query)  # type: ignore
        return DuckDataFrame(result)

    def _other_select(self, dfs: DataFrames, statement: str) -> DataFrame:
        conn = duckdb.connect()
        try:
            for k, v in dfs.items():
                duckdb.from_arrow(v.as_arrow(), connection=conn).create_view(k)
            return ArrowDataFrame(_duck_as_arrow(conn.execute(statement)))
        finally:
            conn.close()

    def _get_table(self, table: str) -> Optional[Dict[str, Any]]:
        if isinstance(self.execution_engine, DuckExecutionEngine):
            # TODO: this is over simplified
            con = self.execution_engine.connection
            qt = quote_name(table, "'")
            if not qt.startswith("'"):
                qt = "'" + qt + "'"
            tables = (
                con.query(
                    "SELECT table_catalog,table_schema,table_name,table_type"
                    f" FROM information_schema.tables WHERE table_name={qt}"
                )
                .to_df()
                .to_dict("records")
            )
            return None if len(tables) == 0 else tables[0]
        else:  # pragma: no cover
            raise NotImplementedError(
                "table_exists can only be used with DuckExecutionEngine"
            )


class DuckExecutionEngine(ExecutionEngine):
    """The execution engine using DuckDB.
    Please read |ExecutionEngineTutorial| to understand this important Fugue concept

    :param conf: |ParamsLikeObject|, read |FugueConfig| to learn Fugue specific options
    :param connection: DuckDB connection
    """

    def __init__(
        self, conf: Any = None, connection: Optional[DuckDBPyConnection] = None
    ):
        super().__init__(conf)
        self._native_engine = NativeExecutionEngine(conf)
        self._con = connection or duckdb.connect()
        self._external_con = connection is not None
        self._context_lock = SerializableRLock()
        self._registered_dfs: Dict[str, DuckDataFrame] = {}

        try:
            for pg in list(self._get_pragmas()):  # transactional
                self._con.execute(pg)

            for ext in self.conf.get(_FUGUE_DUCKDB_EXTENSIONS, "").split(","):
                _ext = ext.strip()
                if _ext != "":
                    self._con.install_extension(_ext)
                    self._con.load_extension(_ext)
        except Exception:
            self.stop()
            raise

    def _get_pragmas(self) -> Iterable[str]:
        for k, v in self.conf.items():
            if k.startswith(_FUGUE_DUCKDB_PRAGMA_CONFIG_PREFIX):
                name = k[len(_FUGUE_DUCKDB_PRAGMA_CONFIG_PREFIX) :]
                assert_or_throw(
                    name.isidentifier(), ValueError(f"{name} is not a valid pragma key")
                )
                value = encode_value_to_expr(v)
                yield f"PRAGMA {name}={value};"

    def stop_engine(self) -> None:
        if not self._external_con:
            self._con.close()

    def __repr__(self) -> str:
        return "DuckExecutionEngine"

    @property
    def is_distributed(self) -> bool:
        return False

    @property
    def connection(self) -> DuckDBPyConnection:
        return self._con

    @property
    def log(self) -> logging.Logger:
        return self._native_engine.log

    @property
    def fs(self) -> FileSystem:
        return self._native_engine.fs

    def create_default_sql_engine(self) -> SQLEngine:
        return DuckDBEngine(self)

    def create_default_map_engine(self) -> MapEngine:
        return PandasMapEngine(self._native_engine)

    def get_current_parallelism(self) -> int:
        return 1

    def to_df(self, df: Any, schema: Any = None) -> DataFrame:
        return _to_duck_df(self, df, schema=schema)

    def repartition(
        self, df: DataFrame, partition_spec: PartitionSpec
    ) -> DataFrame:  # pragma: no cover
        self.log.warning("%s doesn't respect repartition", self)
        return df

    def broadcast(self, df: DataFrame) -> DataFrame:
        return df

    def persist(
        self,
        df: DataFrame,
        lazy: bool = False,
        **kwargs: Any,
    ) -> DataFrame:
        # TODO: we should create DuckDB table, but it has bugs, so can't use by 0.3.1
        if isinstance(df, DuckDataFrame):
            # materialize
            res: DataFrame = ArrowDataFrame(df.as_arrow())
        else:
            res = self.to_df(df)
        res.reset_metadata(df.metadata)
        return res

    def join(
        self,
        df1: DataFrame,
        df2: DataFrame,
        how: str,
        on: Optional[List[str]] = None,
    ) -> DataFrame:
        key_schema, output_schema = get_join_schemas(df1, df2, how=how, on=on)
        t1, t2, t3 = (
            TempTableName(),
            TempTableName(),
            TempTableName(),
        )
        on_fields = " AND ".join(
            f"{t1}.{encode_column_name(k)}={t2}.{encode_column_name(k)}"
            for k in key_schema
        )
        join_type = self._how_to_join(how)
        if how.lower() == "cross":
            select_fields = ",".join(
                f"{t1}.{encode_column_name(k)}"
                if k in df1.schema
                else f"{t2}.{encode_column_name(k)}"
                for k in output_schema.names
            )
            sql = f"SELECT {select_fields} FROM {t1} {join_type} {t2}"
        elif how.lower() == "right_outer":
            select_fields = ",".join(
                f"{t2}.{encode_column_name(k)}"
                if k in df2.schema
                else f"{t1}.{encode_column_name(k)}"
                for k in output_schema.names
            )
            sql = (
                f"SELECT {select_fields} FROM {t2} LEFT OUTER JOIN {t1} ON {on_fields}"
            )
        elif how.lower() == "full_outer":
            select_fields = ",".join(
                f"COALESCE({t1}.{encode_column_name(k)},{t2}.{encode_column_name(k)}) "
                f"AS {encode_column_name(k)}"
                if k in key_schema
                else encode_column_name(k)
                for k in output_schema.names
            )
            sql = f"SELECT {select_fields} FROM {t1} {join_type} {t2} ON {on_fields}"
        elif how.lower() in ["semi", "left_semi"]:
            keys = ",".join(encode_schema_names(key_schema))
            on_fields = " AND ".join(
                f"{t1}.{encode_column_name(k)}={t3}.{encode_column_name(k)}"
                for k in key_schema
            )
            sql = (
                f"SELECT {t1}.* FROM {t1} INNER JOIN (SELECT DISTINCT {keys} "
                f"FROM {t2}) AS {t3} ON {on_fields}"
            )
        elif how.lower() in ["anti", "left_anti"]:
            keys = ",".join(encode_schema_names(key_schema))
            on_fields = " AND ".join(
                f"{t1}.{encode_column_name(k)}={t3}.{encode_column_name(k)}"
                for k in key_schema
            )
            sql = (
                f"SELECT {t1}.* FROM {t1} LEFT OUTER JOIN "
                f"(SELECT DISTINCT {keys}, 1 AS __contain__ FROM {t2}) AS {t3} "
                f"ON {on_fields} WHERE {t3}.__contain__ IS NULL"
            )
        else:
            select_fields = ",".join(
                f"{t1}.{encode_column_name(k)}"
                if k in df1.schema
                else f"{t2}.{encode_column_name(k)}"
                for k in output_schema.names
            )
            sql = f"SELECT {select_fields} FROM {t1} {join_type} {t2} ON {on_fields}"
        return self._sql(sql, {t1.key: df1, t2.key: df2})

    def _how_to_join(self, how: str):
        return how.upper().replace("_", " ") + " JOIN"

    def union(self, df1: DataFrame, df2: DataFrame, distinct: bool = True) -> DataFrame:
        assert_or_throw(
            df1.schema == df2.schema, ValueError(f"{df1.schema} != {df2.schema}")
        )
        if distinct:
            t1, t2 = TempTableName(), TempTableName()
            sql = f"SELECT * FROM {t1} UNION SELECT * FROM {t2}"
            return self._sql(sql, {t1.key: df1, t2.key: df2})
        return DuckDataFrame(
            _to_duck_df(self, df1).native.union(_to_duck_df(self, df2).native)
        )

    def subtract(
        self, df1: DataFrame, df2: DataFrame, distinct: bool = True
    ) -> DataFrame:  # pragma: no cover
        if distinct:
            t1, t2 = TempTableName(), TempTableName()
            sql = f"SELECT * FROM {t1} EXCEPT SELECT * FROM {t2}"
            return self._sql(sql, {t1.key: df1, t2.key: df2})
        return DuckDataFrame(
            _to_duck_df(self, df1).native.except_(_to_duck_df(self, df2).native)
        )

    def intersect(
        self, df1: DataFrame, df2: DataFrame, distinct: bool = True
    ) -> DataFrame:
        if distinct:
            t1, t2 = TempTableName(), TempTableName()
            sql = f"SELECT * FROM {t1} INTERSECT DISTINCT SELECT * FROM {t2}"
            return self._sql(sql, {t1.key: df1, t2.key: df2})
        raise NotImplementedError(
            "DuckDB doesn't have consist behavior on INTERSECT ALL,"
            " so Fugue doesn't support it"
        )

    def distinct(self, df: DataFrame) -> DataFrame:
        rel = _to_duck_df(self, df).native.distinct()
        return DuckDataFrame(rel)

    def dropna(
        self,
        df: DataFrame,
        how: str = "any",
        thresh: int = None,
        subset: List[str] = None,
    ) -> DataFrame:
        schema = df.schema
        if subset is not None:
            schema = schema.extract(subset)
        if how == "all":
            thr = 0
        elif how == "any":
            thr = thresh or len(schema)
        else:  # pragma: no cover
            raise ValueError(f"{how} is not one of any and all")
        cw = [
            f"CASE WHEN {encode_column_name(f)} IS NULL THEN 0 ELSE 1 END"
            for f in schema.names
        ]
        expr = " + ".join(cw) + f" >= {thr}"
        return DuckDataFrame(_to_duck_df(self, df).native.filter(expr))

    def fillna(self, df: DataFrame, value: Any, subset: List[str] = None) -> DataFrame:
        def _build_value_dict(names: List[str]) -> Dict[str, str]:
            if not isinstance(value, dict):
                v = encode_value_to_expr(value)
                return {n: v for n in names}
            else:
                return {n: encode_value_to_expr(value[n]) for n in names}

        names = df.columns
        if isinstance(value, dict):
            # subset should be ignored
            names = list(value.keys())
        elif subset is not None:
            names = list(df.schema.extract(subset).names)
        vd = _build_value_dict(names)
        assert_or_throw(
            all(v != "NULL" for v in vd.values()),
            ValueError("fillna value can not be None or contain None"),
        )
        cols = [
            f"COALESCE({encode_column_name(f)}, {vd[f]}) AS {encode_column_name(f)}"
            if f in names
            else encode_column_name(f)
            for f in df.columns
        ]
        return DuckDataFrame(_to_duck_df(self, df).native.project(", ".join(cols)))

    def sample(
        self,
        df: DataFrame,
        n: Optional[int] = None,
        frac: Optional[float] = None,
        replace: bool = False,
        seed: Optional[int] = None,
    ) -> DataFrame:
        assert_or_throw(
            (n is None and frac is not None and frac >= 0.0)
            or (frac is None and n is not None and n >= 0),
            ValueError(
                f"one and only one of n and frac should be non-negative, {n}, {frac}"
            ),
        )
        tb = TempTableName()
        if frac is not None:
            sql = f"SELECT * FROM {tb} USING SAMPLE bernoulli({frac*100} PERCENT)"
        else:
            sql = f"SELECT * FROM {tb} USING SAMPLE reservoir({n} ROWS)"
        if seed is not None:
            sql += f" REPEATABLE ({seed})"
        return self._sql(sql, {tb.key: df})

    def take(
        self,
        df: DataFrame,
        n: int,
        presort: str,
        na_position: str = "last",
        partition_spec: Optional[PartitionSpec] = None,
    ) -> DataFrame:
        partition_spec = partition_spec or PartitionSpec()
        assert_or_throw(
            isinstance(n, int),
            ValueError("n needs to be an integer"),
        )

        if presort is not None and presort != "":
            _presort = parse_presort_exp(presort)
        else:
            _presort = partition_spec.presort
        tb = TempTableName()

        if len(_presort) == 0:
            if len(partition_spec.partition_by) == 0:
                return DuckDataFrame(_to_duck_df(self, df).native.limit(n))
            cols = ", ".join(encode_schema_names(df.schema))
            pcols = ", ".join(encode_column_names(partition_spec.partition_by))
            sql = (
                f"SELECT *, ROW_NUMBER() OVER (PARTITION BY {pcols}) "
                f"AS __fugue_take_param FROM {tb}"
            )
            sql = f"SELECT {cols} FROM ({sql}) WHERE __fugue_take_param<={n}"
            return self._sql(sql, {tb.key: df})

        sorts: List[str] = []
        for k, v in _presort.items():
            s = encode_column_name(k)
            if not v:
                s += " DESC"
            s += " NULLS FIRST" if na_position == "first" else " NULLS LAST"
            sorts.append(s)
        sort_expr = "ORDER BY " + ", ".join(sorts)

        if len(partition_spec.partition_by) == 0:
            sql = f"SELECT * FROM {tb} {sort_expr} LIMIT {n}"
            return self._sql(sql, {tb.key: df})

        cols = ", ".join(encode_schema_names(df.schema))
        pcols = ", ".join(encode_column_names(partition_spec.partition_by))
        sql = (
            f"SELECT *, ROW_NUMBER() OVER (PARTITION BY {pcols} {sort_expr}) "
            f"AS __fugue_take_param FROM {tb}"
        )
        sql = f"SELECT {cols} FROM ({sql}) WHERE __fugue_take_param<={n}"
        return self._sql(sql, {tb.key: df})

    def load_df(
        self,
        path: Union[str, List[str]],
        format_hint: Any = None,
        columns: Any = None,
        **kwargs: Any,
    ) -> LocalBoundedDataFrame:
        dio = DuckDBIO(self.fs, self.connection)
        return dio.load_df(path, format_hint, columns, **kwargs)

    def save_df(
        self,
        df: DataFrame,
        path: str,
        format_hint: Any = None,
        mode: str = "overwrite",
        partition_spec: Optional[PartitionSpec] = None,
        force_single: bool = False,
        **kwargs: Any,
    ) -> None:
        partition_spec = partition_spec or PartitionSpec()
        if not partition_spec.empty and not force_single:
            kwargs["partition_cols"] = partition_spec.partition_by
        dio = DuckDBIO(self.fs, self.connection)
        dio.save_df(_to_duck_df(self, df), path, format_hint, mode, **kwargs)

    def convert_yield_dataframe(self, df: DataFrame, as_local: bool) -> DataFrame:
        if as_local:
            return df.as_local()
        # self._ctx_count <= 1 means it is either not in a context or in the top
        # leve context where the engine was constructed
        return df.as_local() if self._ctx_count <= 1 and not self._external_con else df

    def _sql(self, sql: str, dfs: Dict[str, DataFrame]) -> DuckDataFrame:
        with self._context_lock:
            df = self.sql_engine.select(
                DataFrames(dfs), StructuredRawSQL.from_expr(sql, dialect=None)
            )
            return DuckDataFrame(df.native)  # type: ignore


def _to_duck_df(
    engine: DuckExecutionEngine, df: Any, schema: Any = None, create_view: bool = False
) -> DuckDataFrame:
    def _gen_duck() -> DuckDataFrame:
        if isinstance(df, DuckDBPyRelation):
            assert_or_throw(
                schema is None,
                ValueError("schema must be None when df is a DuckDBPyRelation"),
            )
            return DuckDataFrame(df)
        if isinstance(df, DataFrame):
            assert_or_throw(
                schema is None,
                ValueError("schema must be None when df is a DataFrame"),
            )
            if isinstance(df, DuckDataFrame):
                return df
            rdf = DuckDataFrame(
                duckdb.from_arrow(df.as_arrow(), connection=engine.connection)
            )
            rdf.reset_metadata(df.metadata if df.has_metadata else None)
            return rdf
        tdf = ArrowDataFrame(df, schema)
        return DuckDataFrame(
            duckdb.from_arrow(tdf.native, connection=engine.connection)
        )

    res = _gen_duck()
    if create_view:
        with engine._context_lock:
            if res.alias not in engine._registered_dfs:
                res.native.create_view(res.alias, replace=True)
                # must hold the reference of the df so the id will not be reused
                engine._registered_dfs[res.alias] = res
    return res



from typing import Any, Dict, Iterable, List, Optional

from fugue.dataframe.dataframe import (
    DataFrame,
    LocalBoundedDataFrame,
    _get_schema_change,
    as_fugue_dataset,
)
from fugue.exceptions import FugueDataFrameOperationError
from triad.utils.assertion import assert_or_throw
from triad.utils.pyarrow import apply_schema


class ArrayDataFrame(LocalBoundedDataFrame):
    """DataFrame that wraps native python 2-dimensional arrays. Please read
    |DataFrameTutorial| to understand the concept

    :param df: 2-dimensional array, iterable of arrays, or
      :class:`~fugue.dataframe.dataframe.DataFrame`
    :param schema: |SchemaLikeObject|

    .. admonition:: Examples

        >>> a = ArrayDataFrame([[0,'a'],[1,'b']],"a:int,b:str")
        >>> b = ArrayDataFrame(a)
    """

    def __init__(self, df: Any = None, schema: Any = None):  # noqa: C901
        if df is None:
            super().__init__(schema)
            self._native = []
        elif isinstance(df, DataFrame):
            if schema is None:
                super().__init__(df.schema)
                self._native = df.as_array(type_safe=False)
            else:
                schema, _ = _get_schema_change(df.schema, schema)
                super().__init__(schema)
                self._native = df.as_array(schema.names, type_safe=False)
        elif isinstance(df, Iterable):
            super().__init__(schema)
            self._native = df if isinstance(df, List) else list(df)
        else:
            raise ValueError(f"{df} is incompatible with ArrayDataFrame")

    @property
    def native(self) -> List[Any]:
        """2-dimensional native python array"""
        return self._native

    @property
    def empty(self) -> bool:
        return self.count() == 0

    def peek_array(self) -> List[Any]:
        self.assert_not_empty()
        return list(self.native[0])

    def count(self) -> int:
        return len(self.native)

    def _drop_cols(self, cols: List[str]) -> DataFrame:
        return ArrayDataFrame(self, self.schema - cols)

    def _select_cols(self, keys: List[Any]) -> DataFrame:
        return ArrayDataFrame(self, self.schema.extract(keys))

    def rename(self, columns: Dict[str, str]) -> DataFrame:
        try:
            schema = self.schema.rename(columns)
        except Exception as e:
            raise FugueDataFrameOperationError from e
        return ArrayDataFrame(self.native, schema)

    def alter_columns(self, columns: Any) -> DataFrame:
        new_schema = self._get_altered_schema(columns)
        if new_schema == self.schema:
            return self
        temp = ArrayDataFrame(self.native, new_schema).as_array(type_safe=True)
        return ArrayDataFrame(temp, new_schema)

    def as_array(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> List[Any]:
        if not type_safe and columns is None:
            return self.native
        return list(self.as_array_iterable(columns, type_safe=type_safe))

    def as_array_iterable(
        self, columns: Optional[List[str]] = None, type_safe: bool = False
    ) -> Iterable[Any]:
        if columns is None:
            pos = []
        else:
            pos = [self.schema.index_of_key(k) for k in columns]
            assert_or_throw(len(pos) > 0, "columns if set must be non empty")
        if not type_safe:
            for item in self._iter_cols(pos):
                yield item
        else:
            sub = self.schema if columns is None else self.schema.extract(columns)
            for item in apply_schema(
                sub.pa_schema,
                self._iter_cols(pos),
                copy=True,
                deep=True,
                str_as_json=True,
            ):
                yield item

    def head(
        self, n: int, columns: Optional[List[str]] = None
    ) -> LocalBoundedDataFrame:
        res = ArrayDataFrame(self.native[:n], self.schema)
        return res if columns is None else res[columns]  # type: ignore

    def _iter_cols(self, pos: List[int]) -> Iterable[List[Any]]:
        if len(pos) == 0:
            for row in self.native:
                yield row
        else:
            for row in self.native:
                yield [row[p] for p in pos]


@as_fugue_dataset.candidate(lambda df, **kwargs: isinstance(df, list), priority=0.9)
def _arr_to_fugue(df: List[Any], **kwargs: Any) -> ArrayDataFrame:
    return ArrayDataFrame(df, **kwargs)
